// All public classes/methods in student's file are provided by autograder and can be used freely
// PASS(), RESULT(res), FAIL(), and CHECK_STDOUT() are also provided by autograder and can be used freely


public class TestCaseTemplate1 {
    public static void main(String[] args) {
        // You can call any public method from any public class from student's file like this.
        // It can have any arguments and return values -- use it like any other method.
        int result = Homework.someStudentMethod();
        int SOME_EXPECTED_RESULT = 83;
        if (res == SOME_EXPECTED_RESULT)
            PASS();
        else
            FAIL();
    }
}