__title__ = "assignment-autograder"
__description__ = "Automatic assignment grading for instructor use in programming courses"
__version__ = "3.3.1"
__author__ = "Stanislav Zmiev"
__author_email__ = "szmiev2000@gmail.com"
__license__ = "GPL-3.0"
