# student_submission is provided by autograder, represents student's module, and can be used freely
# PASS(), RESULT(res), FAIL(), and CHECK_STDOUT() are also provided by autograder and can be used freely


def main():
    # You can call any function from student's file like this.
    # It can have any arguments and return values -- use it like any other function.
    student_submission.some_student_function()

    # If you call this function, the program exits, and then autograder will compare this
    # program's stdout with whatever .txt file (with the same name stem as this test case)
    # you have provided in tests/output directory.
    CHECK_STDOUT()


main()
